module doeels_mod
contains
subroutine doeels (e0, theta, phia, phib, wmin, wmax, dw, comment, comment_size,   &
              layers, neps, nper, name, name_size, thick, epsinf, nos, oscType, osc, osc_size, &
              contrl, wn_array, debug, f_array, wn_array_size)

! ******************************************************************
! *                                                                *
! * compute the classical eels spectrum of an arbitrary plane-     *
! * statified medium made from isotropic materials in specular     *
! * geometry using the dielectric theory of eels.                  *
! *                                                                *
! * e0:             impact energy (eV)                             *
! * theta:          incidence angle (°)                            *
! * phia, phib:     angular apertures of the elliptic detector (°) *
! * wmin, wmax, dw: energy-loss interval and step size (cm**-1)    *
! * comment:        comment lines                                  *
! * comment_size:   number of comment lines                        *
! * layers:         number of layers                               *
! * neps:           number of epsilon values                       *
! * nper:           number of periodic layers                      *
! * name:           layer names                                    *
! * name_size:      number of layers                               *
! * thick:          layer thickness                                *
! * epsinf:         epsilon at infinite frequency                  *
! * nos:            number of oscillators                          *
! * oscType:        oscillator type: Lorentz, Kurosawa, drude      *
! * osc:            oscillator parameters, wTO, Q, lambda          *
! * osc_size:       number of oscillators                          *
! * contrl:         'image' for image-charge screening             *
! * wn_array:       frequencies                                    *
! * debug:          logical                                        *
! * f_array:        spectrum                                       *
! * wn_array_size:  number of frequencies                          *
! *                                                                *
! ******************************************************************

  use quanc8_mod
  use queels_mod
  use seteps_mod
  use param_mod

  implicit none

  integer, parameter :: nt = 5

  double precision, intent(in) :: e0, theta, phia, phib, wmin, wmax, dw
  character (len = 72) :: comment(comment_size)
  character (len = 10) :: name(name_size)
  double precision, intent(in) :: thick(name_size), epsinf(name_size), osc(4, osc_size)
  character (len = 10) :: contrl
  integer, intent(in) :: comment_size, name_size, osc_size, wn_array_size, oscType(:)
  integer, intent(in) :: layers, nper, nos(name_size)
  double precision, intent(in out) :: wn_array(wn_array_size), f_array(wn_array_size)
  logical, intent(in), optional :: debug

  integer :: i, iw, neps, nofu, nout, nw, lmax
  double precision :: a, aerr, alpha, b, beta,              &
      c1, c2, dx, ener, facru, f, f0,                       &
      f1, fpic, fun, pi, prefac, psia, psii, qrat, rerr, t, &
      width, wpic, x, xmin, xmax, z, z1, z2
  double precision :: table(nt)
  double complex, allocatable :: eps(:)

  external fun, qrat

  data aerr / 0.0d0 /, rerr / 1.0d-06 /, f / 0.0d0 /, f1 / 0.0d0 /

  if (debug) then
    write (*,*) 'doeels:'
    write (*,*) 'comment: ',  size(comment)
    write (*,*) 'name: ',     size(name)
    write (*,*) 'thick: ',    size(thick)
    write (*,*) 'epsinf: ',   size(epsinf)
    write (*,*) 'osc: ',      size(osc), size(osc, 1), size(osc, 2)
    write (*,*) 'nos: ',      size(nos)
    write (*,*) 'wn_array: ', size(wn_array)
    write (*,*) 'f_array: ',  size(f_array)
  endif

! *** machine-dependent constants
! *** epsmac + 1.0 = epsmac , cosh(argmin) = 1.0 , tanh(argmax) = 1.0

  pi = 4 * datan(1.0d0)

  dlimf = 0.0d0
  rational = .false.

! *** read target specifications

  user = layers == 0
  if (user) then

    if (layers == 1) then
      rational = .true.
    endif
    if (contrl == 'image') then
! *** image-charge screening factor
      if (layers == 1 .and. neps == 2) then
        dlimf = dsqrt(epsinf(1) * epsinf(2))
      else
        dlimf = epsinf(1)
      endif
      dlimf = (dlimf - 1.0d0) / (dlimf + 1.0d0)
    endif
  endif

! *** initialize constants

  lmax = size(thick)
  nw = size(wn_array)
  if (debug) write (*,*) 'lmax: ', lmax
  allocate(eps(lmax))
  if (debug) write (*,*) 'eps: ', size(eps)
  nout = 1 + nw / 20
  ener = 8065 * e0
  psia = phia / 180 * pi
! *** kms
  psii = theta / 180 * pi
  cospsi = dcos(psii)
  sinpsi = dsin(psii)
  tanpsi = dtan(psii)
! *** f2023 replacement suggestion
!  cospsi = dcosd(phia)
!  sinpsi = dsind(phia)
!  tanpsi = dtand(phia)
  prefac = dsqrt(255500 / e0)/(137 * cospsi)
  facru = psia / cospsi * dsqrt(0.2624664d0 * e0)
  elleps = (1.0d0 - phia / phib) * (1.0d0 + phia / phib)
  acoef = sinpsi**2 + elleps * cospsi**2
  bcoef = sinpsi * cospsi
  if (dlimf > 0.0d0) then
    rational = .false.
    if (debug) write(*,*) ' = > electron attracted by an image charge = ', dlimf
! *** dlimf : half the length unit imposed by the image force
    dlimf = 1.80d0 * dlimf/(e0 * cospsi**2)
  endif
  if (debug) write (*,*) 'rational: ', rational
  if (rational) then

! *** set up coefficients for the rational approximation to the integral

    if (debug) write(*,*) ' = > set up a rational approximation to the integral'
    call quanc8(fun, 0.0d0, pi / 2, aerr, rerr, alpha, c1, nofu, c2, eps, thick, layers, nper)
    alpha  = (2 / pi)**2 * alpha
    c1 = 2 / pi / dsqrt(1.0d0 - elleps) * sinpsi * alpha**2
    if (c1 > 0.99d0) then
      if (debug) write(*,*) ' ===> cannot do it'
      rational = .false.
    else
      c2 = 3 * alpha**2 / (1.0d0 - c1)
      c1 = c1 * c2
      xmin = wmin / (2 * ener * psia)
      xmax = wmax / (2 * ener * psia)
      if (xmin <= 0.0d0) then
        xmin = 0.0d0
      endif
      dx = dmax1(0.02d0, (xmax - xmin) / nt)
      z1 = 0.0d0
      z2 = 0.0d0
      do i = 1, nt
        x = xmin + i * dx
        call queels(x, f, aerr, rerr, facru, eps, thick, layers, nper)
        table(i) = f
        f = f * (1.0d0 + alpha * x)**2
        if (dabs(c2 * f - c1) < c2 * rerr) then
          cycle
        endif
        z = (1.0d0 - f) / (c2 * f - c1)
        if (z <= 0.0d0) then
          cycle
        endif
        z1 = z1 + x * z * (x**2 - z)
        z2 = z2 + (x * z)**2
      enddo
      if (z2 == 0.0d0) then
        if (debug) write(*,*) ' ===> cannot do it'
        rational = .false.
      else
        beta = z1 / z2
        z = 0.0d0
        do i = 1, nt
          x = xmin + i * dx
          z = z + (table(i) - qrat(x, alpha, beta, c1, c2))**2
        enddo
        z = dsqrt(z) / nt
        if (z > 5.0d-03) then
          if (debug) write(*,*) ' ===> cannot do it'
          rational = .false.
        else
          if (debug) write(*, 100) alpha, c1, c2, beta, z
        endif ! z > 5.0d-03
      endif ! z2 == 0.0d0
    endif ! c1 > 0.99d0
  endif ! rational

! *** loop over the energy losses

  if (debug) write(*, 110)
  do iw = 1, nw
    f0 = f1
    f1 = f
    f = 0.0d0
    wn = wmin + (iw - 1) * dw
!    if (debug) write (*,*) 'wn: ', wn
    if (wn >= 0.0d0) then
      if (wn /= 0.0d0) then
        if (.not. user) call seteps(neps, nos, oscType, osc, epsinf, wn, eps, layers)

        x = wn / (2 * ener * psia)
        if (rational) then
          f = qrat(x, alpha, beta, c1, c2) * dimag(-2 / (1.0d0 + eps(1)))
        else
          call queels(x, f, aerr, rerr, facru, eps, thick, layers, nper)
        endif
        f = prefac * f / wn
      endif ! wn /= 0.0d0

      wn_array(iw) = wn
      f_array(iw)  = f

! *** localize a peak using a parabolic interpolation

      if ((iw >= 3) .and. (f1 - f0 > aerr) .and. (f1 - f > aerr)) then
        a = (f1 - f0) + (f1 - f)
        if (a > 4 * rerr * f1) then
          b = 0.5d0 * (f1 - f0 + 3 * (f1 - f))
          t = b / a
          wpic = wn - t * dw
          fpic = f + 0.5d0 * b * t
          width = dsqrt(8 * fpic / a) * dw
          if (debug) write(*, 120) wpic, fpic, width
        endif ! a > 4 * rerr * f1
      endif ! iw >= 3 ...
    endif ! wn >= 0.0d0
    if (mod(iw, nout) == 0) then
      if (debug) write(*, 130) 100 * iw / nw, wn, f
    endif
  enddo
  return
100 format(5x, 'alpha = ', f9.4, 4x, 'c1 = ', f9.4, 4x, 'c2 = ', f9.4, 4x,  &
      'beta = ', f9.4/5x, 'accuracy = ', e9.2)
110 format(//' run (%)  wn (cm**-1)  pcl(wn) (cm) |',  &
      ' peak location  amplitude    width')
120 format(40x, f10.2, d12.4, f10.2)
130 format(2x, f5.1, 3x, f11.3, d14.5)
end subroutine doeels
end module doeels_mod
