program eels_boson

! ******************************************************************
! *                                                                *
! * compute the classical eels spectrum of an arbitrary plane-     *
! * statified medium made from isotropic materials in specular     *
! * geometry using the dielectric theory of eels.                  *
! *                                                                *
! ******************************************************************

  use, intrinsic :: iso_fortran_env, only:  output_unit

  use get_commandline_options_mod
  use doeels_mod
  use oscillatorType_mod
  use getOscillator_mod
  use fortran_yaml_c

  implicit none

  include 'version.inc'

  type(YamlFile), target :: file
  character(:), allocatable :: err
    
  class(type_node), pointer :: root
  type(type_error), allocatable :: io_err
  class(type_list), pointer :: layerList, termList
  class(type_list_item), pointer :: layerListItem, termListItem

  double precision :: e0, theta, phia, phib, wmin, wmax, dw
  integer :: i, j, jos, l, layers, neps, nper, nw
  logical :: user
  character (len = 72) :: comment(2)
  character (len = 10) :: contrl
  double precision, allocatable :: xout(:), yout(:)
  integer :: nout

  double precision, allocatable :: thick(:), epsinf(:), osc(:, :)
  integer, allocatable :: nos(:), oscType(:), tmp_oscType(:)
  character (len = 10), allocatable :: name(:)
  double precision, allocatable :: wn_array(:), f(:)

  integer :: old_size, old_size_1, old_size_2
  double precision, allocatable :: tmp_osc(:, :)

  double precision :: asym, emax, emin, gauss, t, width
  logical :: debug
  
  character (len = :), allocatable :: eelsin_name, bosonout_name

  ! the default file names
  eelsin_name    = 'eelsin.yaml'
  bosonout_name  = 'bosonout.txt'
  
  call get_commandline_options(eelsin_name, bosonout_name)

  call file%parse(eelsin_name, err)
  if (allocated(err)) then
    print*,err
    stop 1
  endif

  call file%dump(unit = output_unit, indent = 0)

  root => file%root
  
  select type (root)
    class is (type_dictionary)
      
! *** assign spectrometer parameters
! impact energy (ev)
  e0 = root%get_real('E0', error = io_err)
  if (allocated(io_err)) then
    print *, io_err%message
    stop 1
  endif
! incidence angle (%)
  theta = root%get_real('Theta', error = io_err)
  if (allocated(io_err)) then
    print *, io_err%message
    stop 1
  endif
! angular apertures of the elliptic detector (%)
  phia = root%get_real('PhiA', error = io_err)
  if (allocated(io_err)) then
    print *, io_err%message
    stop 1
  endif
  phib = root%get_real('PhiB', error = io_err)
  if (allocated(io_err)) then
    print *, io_err%message
    stop 1
  endif
! energy-loss interval and step size (cm**-1)
  wmin = root%get_real('wMin', error = io_err)
  if (allocated(io_err)) then
    print *, io_err%message
    stop 1
  endif
  wmax = root%get_real('wMax', error = io_err)
  if (allocated(io_err)) then
    print *, io_err%message
    stop 1
  endif
  dw = root%get_real('dw', error = io_err)
  if (allocated(io_err)) then
    print *, io_err%message
    stop 1
  endif
! comment lines
  comment(1) = root%get_string("Comment1", error = io_err)
  if (allocated(io_err)) then
    print *, io_err%message
    stop 1
  endif
  comment(2) = root%get_string("Comment2", error = io_err)
  if (allocated(io_err)) then
    print *, io_err%message
    stop 1
  endif
  nper = root%get_integer("NumberOfPeriodicLayers", error = io_err)
  if (allocated(io_err)) then
    print *, io_err%message
    stop 1
  endif


    layerList => root%get_list('Layers', required = .true., error = io_err)
    if (allocated(io_err)) then
      print *, io_err%message
      stop 1
    endif

    print'(1x,a,i3)',"Number of layers: ", layerList%size()
  end select

  write(*,*) 'program eels (September 2020), version: ', version
  write(*,'(a, f6.2, a, f5.1, a, f5.2, a, f5.2, a)') ' e0 = ', e0,       &
    ' eV, theta = ', theta, '°, phia = ', phia, '°, phib = ', phib, '°'
  write(*,'(a, g11.4, a, g11.4, a, g11.4, a)') ' energy losses from',    &
    wmin, ' to', wmax, ', step = ', dw, ' cm**-1'
  write(*,*) comment(1)
  write(*,*) comment(2)

  if (phia <= 0.0d0 .or. phib <= 0.0d0) then
    stop '*** wrong input ***'
  endif
  if (e0 <= 0.0d0 .or. theta + phia >= 90.0d0) then
    stop '*** bad input ***'
  endif

! *** read target specifications

  layers = layerList%size()
  user = layers == 0
  if (.not. user) then
    neps = layers
    if (nper == -1) then
      neps = layers + 1
      nper = 1
      write(*,*) 'The substrate is an anisotropic uniaxial material'
    endif
    if (layers < 0 .or. nper < 1 .or. nper > layers) then
      stop '*** invalid target specifications ***'
    endif
    write(6, 101) layers, nper
    jos = 0
    allocate (name(neps), thick(neps), epsinf(neps), nos(neps))
    layerListItem => layerList%first
    call layerListItem%node%dump(unit = output_unit, indent = 0)

    do l = 1, neps
      if (l <= layers) then
        write (*,*)
        write (*,*) '*** Layer:', l, 'from', layers, 'layers. ***'
        write (*,*)
        select type (layer => layerListItem%node)
        class is (type_dictionary)
          name(l) = layer%get_string('Name', error = io_err)
          if (allocated(io_err)) then
            print *, io_err%message
            stop 1
          endif
          thick(l) = layer%get_real('thickness', error = io_err)
          if (allocated(io_err)) then
            print *, io_err%message
            stop 1
          endif
          epsinf(l) = layer%get_real('epsInfinity', error = io_err)
          if (allocated(io_err)) then
            print *, io_err%message
            stop 1
          endif
          termList => layer%get_list('Terms', required = .true., error = io_err)
          if (allocated(io_err)) then
            print *, io_err%message
            stop 1
          endif
          nos(l) = termList%size()
        end select
        write (*,*) 'Name: ', name(l)
        write (*,*) 'Thickness: ', thick(l)
        write (*,*) 'epsInfinity: ', epsinf(l)
        write (*,*) 'Number of Oscillators: ', nos(l)
      endif
      write(6, 103)
      if (nos(l) <= 0) then
        if (l <= layers) write(6, 104) l, name(l), thick(l), epsinf(l)
        if (l  > layers) write(6, 105) epsinf(l)
      else
        termListItem => termList%first
        do j = 1, nos(l)
          if (.not. allocated(osc)) then
            allocate(osc(4, nos(l)))
          else if (j == 1) then
            old_size_1 = size(osc, 1)
            old_size_2 = size(osc, 2)
            allocate(tmp_osc(old_size_1, old_size_2 + nos(l)))
            tmp_osc(1:old_size_1, 1:old_size_2) = osc
            deallocate(osc)
            call move_alloc(tmp_osc, osc)
          endif
          if (.not. allocated(oscType)) then
            allocate(oscType(nos(l)))
          else if (j == 1) then
            old_size = size(oscType)
            allocate(tmp_oscType(old_size + nos(l)))
            tmp_oscType(1:old_size) = oscType
            deallocate(oscType)
            call move_alloc(tmp_oscType, oscType)
          endif
          jos = jos + 1
          call getOscillator(termListItem,  oscType, osc, jos)
          if ((j == nos(l) / 2 + 1) .and. (nos(l) > 1)) then
            write(6, 106)
            write(6, 107)
          endif
          if (j == 1) then
            if (l <= layers) then
              write(6, 104) l, name(l), thick(l), epsinf(l), (osc(i, jos), i = 1, 4)
            else
              write(6, 105) epsinf(l), (osc(i, jos), i = 1, 4)
            endif
          else
            write(6, 108) (osc(i, jos), i = 1, 4)
          endif
          termListItem => termListItem%next
        enddo
      endif
      layerListItem => layerListItem%next
    enddo
    write(*,*)
  endif

  nw = 1 + int((wmax - wmin) / dw)
  allocate (wn_array(nw), f(nw))

  debug = .false.
  call doeels(e0, theta, phia, phib, wmin, wmax, dw, comment, size(comment),   &
              layers, neps, nper, name, size(name), thick, epsinf, nos, oscType, osc, size(osc, 2),&
              contrl, wn_array, debug, f, size(wn_array))

  select type (root)
    class is (type_dictionary)
! target temperature (Kelvin)
    t = root%get_real('T', error = io_err)
    if (allocated(io_err)) then
      print *, io_err%message
      stop 1
    endif
! width of the instrumental response (cm**-1)
    width = root%get_real('Width', error = io_err)
    if (allocated(io_err)) then
      print *, io_err%message
      stop 1
    endif
! fraction of gaussian for the instrumental response
    gauss = root%get_real('Gauss', error = io_err)
    if (allocated(io_err)) then
      print *, io_err%message
      stop 1
    endif
! asymmetry of the instrumental response
    asym = root%get_real('Asym', error = io_err)
    if (allocated(io_err)) then
      print *, io_err%message
      stop 1
    endif
! lower and upper energy losses for this computation (cm**-1)
    emin = root%get_real('EMin', error = io_err)
    if (allocated(io_err)) then
      print *, io_err%message
      stop 1
    endif
    emax = root%get_real('EMax', error = io_err)
    if (allocated(io_err)) then
      print *, io_err%message
      stop 1
    endif
  end select

  write(*,*) 'program boson (September 2020)'
  write(*,'(a, f6.1, a, f7.2, a)') ' t =', t, ' K, width =', width, ' cm**-1'
  write(*,'(a, f5.2, a, f5.2)') ' gauss =', gauss, ', asym =', asym
  write(*,'(a, g11.4, a, g11.4, a)') ' energy losses from', emin, ' to', emax, ' cm**-1'
  write(*,*) comment(2)
  write(*,'(a, i6, a, g15.7, a, g15.7)') ' read', nw, ' values of pcl from', wmin, ' to', wmax

! length calculation for xout, yout.
  nout = nint((emax - emin) / dw ) + 1
  allocate (xout(nout))
  allocate (yout(nout))
  
  debug = .false.

  call doboson(t, width, gauss, asym, emin, emax, &
               wmin, wmax, nw, f, debug, xout, yout, nout)

  comment(2) = comment(1)
  open(unit = 14, file = bosonout_name)
  write(14, '(a, f6.2, a, f5.1, a, f5.2, a, f5.2, a, f6.1, a, f5.2)') &
        'e0 =', e0, ' theta =', theta, ' phia =', phia, ' phib =', phib, ' T =', t, ' GAUSS =', gauss
  write(14, *) comment(2)
  do i = 1, nout
    write(14, '(2e15.7)') xout(i), yout(i)
  end do
  close(unit = 14)
  write(*,*) nout, ' values written on disk'

  deallocate (xout, yout)

  stop 'Successful end!'

101 format(i3, ' layer(s), nper = ', i2//'   l', 2x, 'material', 7x,  &
      'thickness', 5x, 'epsinf', 4x, 'wto , wp', 5x, 'q', 7x, 'gam/wto')
103 format(1x, 72('-'))
104 format(1x, i3, 2x, a10, g15.3, f10.4, f12.4, f10.4, f9.4, f9.4)
105 format(31x, f10.4, f12.4, f10.4, f9.4, f9.4)
106 format(45x, 'wlo , wp', 5x, 'q', 7x, 'gam/wlo')
107 format(45x, 28('-'))
108 format(41x, f12.4, f10.4, f9.4, f9.4)
end program eels_boson
