module queels_mod
contains
subroutine queels(x, f, aerr, rerr, facru, eps, thick, layers, nper)

! ******************************************************************
! *                                                                *
! * perform q-space integration for computing the eels spectrum of *
! * a isotropic target using polar coordinates.                    *
! *                                                                *
! * x is the dimensionless energy loss hbar*omega/(2*e0*phia)      *
! * aerr and rerr are the desired absolute and relative accuracies *
! * facru*x is the units of wavevectors omega/v_perpendicular      *
! * f is the q-integral multiplied by (2/pi)**2                    *
! *                                                                *
! * eps:    epsilon                                                *
! * thick:  thickness                                              *
! * layers: number of layers                                       *
! * nper:   number of periodic layers                              *
! *                                                                *
! ******************************************************************

  use quanc8_mod
  use fint1_mod
  use fint2_mod
  use fint3_mod
  use param_mod

  implicit none

  double precision, intent(in) :: x
  double precision, intent(out) :: f
  double precision, intent(in out) :: aerr
  double precision, intent(in out) :: rerr
  double precision, intent(in) :: facru
  double complex, intent(in) :: eps(:)
  double precision, intent(in) :: thick(:)
  integer, intent(in) :: layers, nper

  double precision :: error, flag
  double precision :: u1, u2, ut, y
  integer :: ie, nofu
  dimension error(3), flag(3)

!  write (*,*) 'queels:'
!  write (*,*) 'eps: ', size(eps)
!  write (*,*) 'thick: ', size(thick)

  f = 0.0d0
  if (x <= 0.0d0) then
    return
  endif
  ru = facru * x
  ccoef = cospsi**2 / x
  ut = ccoef - bcoef
  u1 = dabs(ut)
  u2 = ccoef + bcoef
  if (ut > 0.0d0) then
    call quanc8(fint1, 0.0d0, u1, aerr, rerr, y, error(1), nofu, flag(1), eps, thick, layers, nper)
    f = y
  else
    flag(1) = 0.0d0
  endif
  if (u2 > u1) then
    call quanc8(fint2, u1, u2, aerr, rerr, y, error(2), nofu, flag(2), eps, thick, layers, nper)
    f = f + y
  else
    flag(2) = 0.0d0
  endif
  if (dabs(acoef) > x * (1.0d0 - elleps) * bcoef) then
    um = dsqrt(ccoef / x / (1.0d0 - elleps) + bcoef**2 / acoef)
    if (um > u2) then
      call quanc8(fint3, u2, um, aerr, rerr, y, error(3), nofu, flag(3), eps, thick, layers, nper)
      f = f + y
    endif
    if (um < u1) then
      call quanc8(fint3, um, u1, aerr, rerr, y, error(3), nofu, flag(3), eps, thick, layers, nper)
      f = f - y
    endif
  else
    flag(3) = 0.0d0
  endif
  do ie = 1, 3
    if (flag(ie) == 0.0d0) then
      cycle
    endif
    write(*,*) ' +++ flag(', ie, ') =', flag(ie), ', error =', error(ie), ' +++'
    if (flag(ie) - aint(flag(ie)) > 0.5d-02) then
      stop '*** execution aborted ***'
    endif
  enddo
  f = (2 / 3.141592653589793238d0)**2 * f
  return
end subroutine queels
end module queels_mod
