module rcffi_mod
contains
subroutine rcffi(ar, ai, msign, h)

!   *******************************************************************
!   *                                                                 *
!   * using a radix-two fast-fourier-transform technique, rcffi       *
!   * computes the fourier integral transform of a real function      *
!   * f(x) or the inverse fourier transform of a complex function     *
!   * such that g(-y) = conj(g(y)).                                   *
!   * (adapted from cfft routine, cern library member d704)           *
!   *                                                                 *
!   * g(y) = integral f(x) * cexp(-i * x * y) dx    (msign < 0)       *
!   *                                                                 *
!   * f(x) = 1 / (2 * pi) integral g(y) * cexp(+i * y * x) dy  (msign > 0)    *
!   *                                                                 *
!   * f(x) is tabulated on the meshes of points defined by            *
!   * xj = j * hx and -xj = -j * hx,  j = 0, +1, ..., n-2, n-1        *
!   * ar(j+1) contains f(+xj) and ai(j+1) contains f(-xj)             *
!   * (input when msign < 0 , output when msign >0)                   *
!   * ar(1) may be different from ai(1)                               *
!   *                                                                 *
!   * g(y) is tabulated on the mesh of points defined by              *
!   * yk = k * hy  , k = 0, 1, ..., n - 2, n - 1                      *
!   * ar(j+1) contains real(g(yk)) and ai(j+1) contains aimag(g(yk))  *
!   * (output when msign > 0 , input when msign < 0)                  *
!   * remark :  g(-yk) = conj(g(+yk))                                 *
!   *                                                                 *
!   * n = 2**iabs(msign)  (msign is an input)                         *
!   *                                                                 *
!   * the step sizes satisfy                                          *
!   * hy = (2 * pi) / (n * hx)   or    hx = (2 * pi) / (n * hy)       *
!   * h is an input (h = hx if msign < 0 , h = hy if msign > 0)       *
!   * h must be positive                                              *
!   *                                                                 *
!   * ar and ai are two real arrays with dimension n (input and       *
!   * output)                                                         *
!   *                                                                 *
!   *******************************************************************

  implicit none

  double precision, intent(in out) :: ar(*)
  double precision, intent(in out) :: ai(*)
  integer, intent(in)  :: msign
  double precision, intent(in) :: h

  double precision :: a, as, c, s, t, ti, tr, ui, ur, wi, wr
  integer :: i, ip, j, k, l, le, le1, m, n, nm1, nv2
  logical :: debug

  debug = .false.
  as = 0.0d0
  if (msign == 0) then
    return
  endif

  if (h <= 0.0d0) then
    if (debug) write(*,*) ' *** negative step size in <rcffi>, h = ', h, ' *** '
    stop
  endif

! *** initialization

  m = iabs(msign)
  n = 2**m
  if (msign > 0) then
    ar(1) = 0.5d0 * ar(1)
  else
    as = ar(1) - ai(1)
    ar(1) = 0.5d0 * (ar(1) + ai(1))
    do i = 2, n
      ar(i) = ar(i) + ai(n - i + 2)
    enddo
    do i = 1, n
      ai(i) = 0.0d0
    enddo
  endif
! *** discrete fast-fourier transform
  nv2 = n / 2
  nm1 = n - 1
  j = 1
  do i = 1, nm1
    if (i < j) then
      tr = ar(j)
      ar(j) = ar(i)
      ar(i) = tr
      ti = ai(j)
      ai(j) = ai(i)
      ai(i) = ti
    endif
    k = nv2
    do while (j > k)
      j = j - k
      k = k / 2
    enddo
    j = j + k
  enddo
  do i = 1, n, 2
    tr = ar(i + 1)
    ar(i + 1) = ar(i) - tr
    ar(i) = ar(i) + tr
    ti = ai(i + 1)
    ai(i + 1) = ai(i) - ti
    ai(i) = ai(i) + ti
  enddo
  if (m == 1) then
    return
  endif
  c = 0.0d0
  s = dble(isign(1, msign))
  le = 2
  do l = 2, m
    wr = c
    ur = wr
    wi = s
    ui = wi
    c = dsqrt(c * 0.5d0 + 0.5d0)
    s = wi / (c + c)
    le1 = le
    le = le1 + le1
    do i = 1, n, le
      ip = i + le1
      tr = ar(ip)
      ar(ip) = ar(i) - tr
      ar(i)  = ar(i) + tr
      ti = ai(ip)
      ai(ip) = ai(i) - ti
      ai(i)  = ai(i) + ti
    enddo
    do j = 2, le1
      do i = j, n, le
        ip = i + le1
        tr = ar(ip) * ur - ai(ip) * ui
        ti = ar(ip) * ui + ai(ip) * ur
        ar(ip) = ar(i) - tr
        ar(i)  = ar(i) + tr
        ai(ip) = ai(i) - ti
        ai(i)  = ai(i) + ti
      enddo
      tr = ur
      ur = tr * wr - ui * wi
      ui = tr * wi + ui * wr
    enddo
  enddo

! *** correction of the discrete fft, using a trapezoidal approximation
! *** for the variations of the input function

  c = h
  if (msign >= 0) then
    c = c / 3.141592653589793238d0
    ai(1) = ar(1)
    do i = 2, n
      ai(i) = ar(n - i + 2)
    enddo
  endif
  ar(1) = c * ar(1)
  ai(1) = c * ai(1)
  a = 3.141592653589793238d0 / n
  wr = dcos(a)
  wi = dsin(a)
  t = dsqrt(c)
  a = a / t
  c = 0.0d0
  ur = 1.0d0
  ui = 0.0d0
  do i = 2, n
    c = c + a
    tr = ur
    ur = wr * tr - wi * ui
    ui = wi * tr + wr * ui
    ti = ui / c
    tr = ti**2
    ar(i) = ar(i) * tr
    ai(i) = ai(i) * tr
    if (msign <= 0) then
      if (as /= 0.0d0) then
        ai(i) = ai(i) - as * (t - ur * ti) / (2 * c)
      endif
    endif
  enddo
  return
end subroutine rcffi
end module rcffi_mod
