module seteps_mod
contains
subroutine seteps(neps, nos, oscType, osc, epsinf, wn, eps, layers)

! ******************************************************************
! *                                                                *
! * set up long-wavelength dielectric functions of the layers for  *
! * the present frequency wn (in cm**-1)                           *
! *                                                                *
! * neps:           number of epsilon values                       *
! * nos:            number of oscillators                          *
! * oscType:        oscillator type: Lorentz, Kurosawa, drude      *
! * osc:            oscillator parameters, wTO, Q, lambda          *
! * epsinf:         epsilon at infinite frequency                  *
! * wn:             frequency                                      *
! * eps:            epsilon                                        *
! * layers:         number of layers                               *
! *                                                                *
! ******************************************************************

  use oscillatorType_mod
  
  implicit none

  integer, intent(in) :: neps, nos(:), oscType(:)
  double precision, intent(in) :: osc(:, :), epsinf(:), wn
  double complex, intent(in out) :: eps(:)
  integer, intent(in) :: layers

  integer :: i, j, k
  double complex :: one, nomi, deno, addLorentz, addDrude
  double precision :: wn2
  double precision :: wTO, wLO, wP, Qk, yk, yTO, yLO, y, y0, yp

  one = dcmplx(1.0, 0.0)
  wn2 = wn**2
  k = 1
  do i = 1, neps
    nomi = one
    deno = one
    addLorentz = dcmplx(0.0, 0.0)
    addDrude = dcmplx(0.0, 0.0)
    do j = 1, nos(i)
      select case (oscType(k))
        case (Lorentz) ! Lambin 1990, equ. 7
          wTO = osc(1,k)
          Qk  = osc(2,k)
          yk  = osc(3,k)
          addLorentz = addLorentz + Qk * wTO**2 / dcmplx(wTO**2 - wn2, -yk * wn)
        case (Kurosawa) ! Gervais 2002, equ. 2.16
          wTO = osc(1,k)
          yTO = osc(2,k)
          wLO = osc(3,k)
          yLO = osc(4,k)
          nomi = nomi * dcmplx(wLO**2 - wn2, -yLO * wn)
          deno = deno * dcmplx(wTO**2 - wn2, -yTO * wn)
        case (Drude)! Lambin 1990, equ. 8
          wp = osc(1,k)
          y  = osc(2,k)
          addDrude = addDrude - wp**2 / dcmplx(wn2, y * wn)
        case (extendedDrude) ! Gervais 2002, equ. 2.14
          wp = osc(1,k)
          y0 = osc(2,k)
          yp = osc(3,k)
          addDrude = addDrude - dcmplx(wp**2, (y0 - yp) * wn) / dcmplx(wn2, wn * y0)
        case default
          write (*,'(A)') '*** Error in oscillator type! ***'
      end select
      eps(i) = epsinf(i) * (nomi / deno + addDrude) + addLorentz
      k = k + 1
    enddo
  enddo

  if (neps == layers + 1) then
! the substrate is a anisotropic uniaxial material
    eps(layers) = cdsqrt(eps(layers) * eps(layers + 1))
    if (dimag(eps(layers)) < 0.0d0) then
      eps(layers) = -eps(layers)
    endif
  endif
  return
end subroutine seteps
end module seteps_mod
