module surlos_mod
contains

logical function zero(z)
  double complex, intent(in) :: z
  zero = (dble(z) == 0.0d0) .and. (dimag(z) == 0.0d0)
end function zero  

double precision function surlos(dk, eps, thick, layers, nper)

! ******************************************************************
! *                                                                *
! * eels surface loss function for an arbitrary multilayered target*
! *                                                                *
! * dk:                                                            *
! * eps:    epsilon                                                *
! * thick:  thickness                                              *
! * layers: number of layers                                       *
! * nper:   number of periodic layers                              *
! *                                                                *
! ******************************************************************

  implicit none

  double precision, intent(in) :: dk
  double complex, intent(in) :: eps(:)
  double precision, intent(in) :: thick(:)
  integer, intent(in) :: layers, nper

  integer :: lmax
  logical :: static, skip
  integer :: lstart, n
  double precision, allocatable :: arg(:)
  double precision :: argmin, argmax, cn, cnm1, epsmac, sn, snm1, tn
  double complex :: a, b, csi, pnm2, pnm1, pn, pp, qnm2, qnm1, qn, qp

  epsmac = epsilon(1.0d0)
  argmin = dsqrt(2 * epsmac)
  epsmac = epsmac / 2
  argmax = 0.5d0 * dlog(2 / epsmac)

!  write (*,*) 'surlos:'
!  write (*,*) 'thick: ', size(thick)
!  write (*,*) 'eps: ', size(eps)

  lmax = size(eps)
  allocate (arg(lmax))
  lstart = layers - nper + 1
  static = .true.
  skip = .false.

  do n = 1, layers
    arg(n) = dk * thick(n)
    if (arg(n) > argmax .or. zero(eps(n))) then
      csi  = eps(n)
      skip = .true.
      exit
    endif
    static = .not. (n >= lstart .and. arg(n) > argmin)
  enddo

  if (.not. skip) then
! *** periodic continued fraction, period = nper

    do ! Do loop is only for the option to skip out of it.
      if (nper <= 1) then
        csi = eps(layers)
      else
        if (static) then
          pn = 0.0d0
          qn = 0.0d0
          do n = lstart, layers
            pn = pn + thick(n) * eps(n)
            qn = qn + thick(n) / eps(n)
          enddo
          if (zero(qn)) then
            n = lstart
            if (n <= 1) then
              surlos = 0.0d0
              return
            endif
            n = n - 1
            csi = eps(n) / dtanh(arg(n))
            exit
          endif ! zero(qn)
          csi = cdsqrt(pn / qn)
          if ((dimag(csi) < 0.0d0) .and. (dble(qn) < 0.0d0)) then
            csi = -csi
          endif
        else ! static
          cn = dcosh(arg(lstart))
          sn = dsinh(arg(lstart))
          pnm1 = 1.0d0
          pn = cn
          pp = eps(lstart) * sn
          qnm1 = 0.0d0
          qn = sn / eps(lstart)
          qp = pn
          do  n = lstart + 1, layers
            cnm1 = cn
            snm1 = sn
            cn = dcosh(arg(n))
            sn = dsinh(arg(n))
            a = eps(n) * sn
            pp = cn * pp + a * pn
            qp = cn * qp + a * qn
            b = (eps(n - 1) / eps(n)) * (sn / snm1)
            a = cnm1 * b + cn
            pnm2 = pnm1
            pnm1 = pn
            qnm2 = qnm1
            qnm1 = qn
            pn = a * pnm1 - b * pnm2
            qn = a * qnm1 - b * qnm2
          enddo
          if (zero(qn)) then
            a = qp - pn
            if (zero(a)) then
              n = lstart
              if (n <= 1) then
                surlos = 0.0d0
                return
              endif
              n = n - 1
              csi = eps(n) / dtanh(arg(n))
              exit
            endif
            csi = pp / a
          else
            a = 0.5d0 * (pn - qp) / qn
            b = cdsqrt(a**2 + pp / qn)
            pn = a - pn / qn
            if (cdabs(pn + b) > cdabs(pn - b)) then
              b = -b
            endif
            csi = a + b
          endif ! zero(qn)
        endif ! static
! *** small-dk limit of the periodic tail

      endif ! nper <= 1

      n = lstart

      exit
    enddo
  endif ! .not. skip

! *** backward algorithm

  do
    n = n - 1
    if (n <= 0) then
      a = csi + 1.0d0
      if (zero(a)) then
        surlos = 2 / epsmac
      else
        surlos = dimag(-2 / a)
      endif
      return
    endif ! n <= 0

    if (arg(n) /= 0.0d0) then
      tn = dtanh(arg(n))
      b = eps(n) + csi * tn
      if (zero(b)) then
        surlos = 0.0d0
        return
      endif
      csi = eps(n) * (csi + tn * eps(n)) / b
    endif
  enddo
end function surlos
end module surlos_mod
